/*
 *  Name: asc2rul3.c
 *  Author: Vincent Lefevre
 *  Version: 1.01
 *  Date: 05 / 04 / 1996
 *
 *  ASC1 --> RUL3 compiler
 *
 *  This compiler comes with the registered version of TETRHEX.
 *  It is NOT freeware and musn't be distributed.
 *
 *  This compiler  is provided  "as is"  without warranty of any kind.  Note
 *  that some bugs probably remain  since I haven't tested all the described
 *  features; yet the provided rule files have been successfully compiled.
 *
 *  If no argument is given,  the ASCII file is read from the standard input
 *  and  the rule file is written to the standard output.  If an argument is
 *  given, it is the name of the ASCII file; the rule file is written to the
 *  file having the same name, but whose extension is "RUL".
 *
 *  WARNING! Compile so that a text stream behaves like a binary stream (the
 *  newline character '\n'  should be written  to a text stream  without any
 *  conversion), because stdout is always a text stream. However there is no
 *  problem if an argument (source file name) is given since the target file
 *  is opened as a binary stream.
 *
 *  Notes:
 *    . If the size of short is not 2  or  if a short  must be aligned on an
 *      even address, SSH2 mustn't be defined.
 *    . The system must be big-endian.
 *    . With most of 32 bits processors  (e.g. ARM, Sparc),  ALIGN4  must be
 *      defined (compile with option -DALIGN4).
 *    . Change the directory separator (DS) if necessary.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SSH2        /* Size of short: 2 on Atari. */
#define DS     '\\' /* Directory separator. */
#define NR     53   /* Number of reserved identifiers. */
#define MAXLEN 256  /* MAXLEN >= 9. */
#define IDMLEN 63   /* Number of characters that are taken *
                     * into account in an identifier.      */
#define NAMLEN 24   /* Length of the rule name buffer. */
#define PLNLEN 24   /* Length of a player name buffer. */
#define TBLEN 4096  /* Length of a text block (help), 2^n. */
#define RCMD_ADD 1024

#define get_num() get_const(1, 0)
#define get_language() get_const(0, 0)

enum error
  { ERR_USAGE = 1, ERR_TOOLNG, ERR_OPEN, ERR_CREATE, ERR_WRITE, ERR_SRCE,
    ERR_EOF, ERR_UNGETC, ERR_SECTION, ERR_SECTNAME, ERR_HEADER, ERR_HDRDEF,
    ERR_SYNTAX, ERR_EXPR, ERR_NOCONST, ERR_MAXNSC, ERR_DEFIDF, ERR_UNDEFIDF,
    ERR_NOIDF, ERR_UNDEFADDRID, ERR_ADDRID, ERR_LABEL, ERR_STRING,
    ERR_STRLONG, ERR_SELECTED, ERR_MEM, ERR_SHP1, ERR_SHP2, ERR_TMANYPS,
    ERR_TMANYSIG, ERR_SIGNAME, ERR_TMANYNBR, ERR_NUMBER, ERR_TEXTPOS,
    ERR_DEFNBR, ERR_HELP, ERR_TMANYSC, ERR_SCORES, ERR_BADNHEX, ERR_LONGIDF,
    ERR_VAR, ERR_DEFINED, ERR_DIV0, ERR_MOD, ERR_BADSIZE, ERR_BADCST,
    ERR_VARIND, ERR_NOVAR, ERR_CMD, ERR_RULPART, ERR_NOPATTERN, ERR_NOSHAPE,
    ERR_UNDEFLBL, ERR_OFFSET };

typedef struct idf
{ struct idf *next;
  long value;
  char idf[IDMLEN+1];
  short var;
}  Idf;

typedef struct idfaddr
{ struct idfaddr *next;
  long value;
  char idf[IDMLEN+1];
  int addr;
  short size;
}  Idfaddr;

typedef struct idlabel
{ struct idlabel *next;
  int addr;
  char idf[IDMLEN+1];
  short size;
}  Idlabel;

typedef struct patshp
{ struct patshp *next;
  short selected;
  char idf[IDMLEN+1];
}  PatShp;

typedef struct shape
{ PatShp patshp;
  char nhex;
  char turnmode;
  long shape[6];
}  Shape;

typedef struct
{ long name;
  long key;
}  Sig;

typedef struct textnum
{ struct textnum *next;
  char len;
  char language;
  char text[128];
}  Textnum;

typedef struct
{ char data[17];
  char numdef;
  Textnum *text;
}  Nbr;

typedef struct textblock
{ struct textblock *next;
  char text[TBLEN];
}  Textblock;

typedef struct help
{ struct help *next;
  long len;
  Textblock *tb;
}  Help;

typedef struct score
{ struct score *next;
  short nhex;
  char name[24];
  long scor;
  long date;
}  Score;

void compfile(void);
void get_cmd(void);
void sto_const(long);
void err_char(char);
void new_idbra(short, short);
void def_idbra(short);
void chg_value(short);
void def_idaddr(short);
Idf *defid0(Idf **, int);
Idf *defidf(void);
void alrdy_def(void);
long defexpr(short);
void syntax_error(void);
long varexpr(short);
void check_varind(long);
long infix(short, short);
void shift_rul(int, char);
short patshp(PatShp **, PatShp **, long, char, short *);
void addpatshp(short *, PatShp *, PatShp **);
short get_string(char *, short);
long get_const(short, short);
void set_noconst(short);
void undef_idf(int, char *);
void read_idf1(short);
void read_idf2(void);
void read_idf3(void);
void error(int, char *);
void insuf_mem(void);
void bad_section_name(void);
void expect(char *);
int readc(void);
int nextc(void);
void prevc(void);
void add(char);
void rulput(char);

int n, c, c2;
short npat = 0, nshp = 0, npos = 0, nsig = 0, nnbr = 0;
short varflag, noconst, ncop, pflag = 0;
long nnrt = 2, textlen = 0, helplen;
FILE *srce = stdin, *dest = stdout;
char *gfn = NULL, buffer[8160];
Idf *idnum = NULL;
Idfaddr *idaddr = NULL;
Idlabel *idl = NULL;
static char parts[7] = { '0', 'A', 'B', 'C', 'P', 'Q', 'S' };
static int paddr[7] = { -1, -1, -1, -1, -1, -1, -1 };
static short hdrdef[14] = { 0 };
long hdrval[13];
char rulname[NAMLEN];
char str_too_long[] = "string too long";
char str_inf[] = "syntax: ( <expr> )";
char div0[] = "division by zero";
char mod[] = "x %% y: y must be positive";
char var_error[] = "the left expression must be a variable";
char str_undef[] = "undefined identifier ";
PatShp *pat = NULL, *shp = NULL;
PatShp *spat[255], *sshp[255];
Sig signals[255];
Nbr numbers[8];
short ntlen[8];
Help *hlp = NULL, *hlplast;
char no_long_idf = 0;
Textblock **thlp;
int nhlp;
Score *scores = NULL, *scolast;
unsigned char nsco[4] = { 0, 0, 0, 0 };
char *rulcmd = NULL;
int rulsize = 0, rullen = 0;

enum RULCODES
  { RUL_ADD = 0x40, RUL_SUB, RUL_MUL, RUL_DIV, RUL_ASL, RUL_ASR, RUL_IFS,
    RUL_MOD, RUL_CST16, RUL_CST32, RUL_EQ, RUL_NE, RUL_LT, RUL_GT, RUL_LE,
    RUL_GE, RUL_FCT, RUL_RTN, RUL_BRA, RUL_IFL, RUL_NOT, RUL_AND, RUL_OR,
    RUL_XOR, RUL_RAND16, RUL_RAND32, RUL_VAR, RUL_STO, RUL_NEXT, RUL_PAUSE,
    RUL_TIMER, RUL_SPEED, RUL_ARRAY16, RUL_ARRAY32, RUL_CASE, RUL_PUSH,
    RUL_READ, RUL_PULL, RUL_ABS, RUL_SQ, RUL_RAND15, RUL_RAND31, RUL_MIN,
    RUL_MAX, RUL_CMP2, RUL_MSG, RUL_REMOVE, RUL_REMOVELOOP, RUL_NHEX,
    RUL_CH2X, RUL_X2H, RUL_X2C, RUL_SETHN, RUL_SETEN, RUL_SETHA, RUL_SETEA,
    RUL_TST, RUL_CHG, RUL_CLR, RUL_SET, RUL_HEX, RUL_PAT, RUL_COL, RUL_NBOR };

enum RESIDF
  { ARRW = 1, ARRL, CASE, IFS, IFL, GOTO, TIMER_ON, TCHG, TCLR, TSET, CHGALL,
    CLRALL, SETALL, RETURN, ABS, CH2X, CHG, CLR, COL, HEX, HEXRF, MAX, MIN,
    MSG, NEXT, NHEX, NSHAPE, PAT, PAUSE, PULL, PUSH, RADAR, RAND15, RAND16,
    RAND31, RAND32, READ, REMOVE, REMOVE_LOOP, SET, SETEA, SETEN, SETHA,
    SETHN, SPEED, SQ, STO, TIMER, TIMER_OFF, TST, VAR, X2C, X2H };

char *residf[NR] =
  { "ARRAY.W", "ARRAY.L", "CASE", "IF.S", "IF", "GOTO", "timer_on", "tchg",
    "tclr", "tset", "chgall", "clrall", "setall", "RETURN", "abs", "ch2x",
    "chg", "clr", "col", "hex", "hexrf", "max", "min", "msg", "next", "nhex",
    "nshape", "pat", "pause?", "pull", "push", "radar?", "rand15", "rand16",
    "rand31", "rand32", "read", "remove", "remove_loop", "set", "setea",
    "seten", "setha", "sethn", "speed", "sq", "sto", "timer", "timer_off",
    "tst", "var", "x2c", "x2h" };

short idfcod[] =
  { RUL_RTN, RUL_ABS, RUL_CH2X, RUL_CHG, RUL_CLR, RUL_COL, RUL_HEX,
    RUL_HEX | 0xC000, RUL_MAX, RUL_MIN, RUL_MSG, RUL_NEXT, RUL_NHEX,
    RUL_HEX | 0xBF00, RUL_PAT, RUL_PAUSE, RUL_PULL, RUL_PUSH,
    RUL_NEXT | 0xBF00, RUL_RAND15, RUL_RAND16, RUL_RAND31, RUL_RAND32,
    RUL_READ, RUL_REMOVE, RUL_REMOVELOOP, RUL_SET, RUL_SETEA, RUL_SETEN,
    RUL_SETHA, RUL_SETHN, RUL_SPEED, RUL_SQ, RUL_STO, RUL_TIMER,
    RUL_TIMER | 0xBF00, RUL_TST, RUL_VAR, RUL_X2C, RUL_X2H };

#ifdef ALIGN4
  long temp;
  #define STO_LONG(w) temp = w; *s++ = (char) (temp >> 24); *s++ = (char) (temp >> 16); *s++ = (char) (temp >> 8); *s++ = (char) temp;
#else
  #define STO_LONG(w) *((long *) s)++ = w;
#endif

main(int argc, char *argv[])
{
  char fname[MAXLEN], *s, *t;
  long *p, tmp;
  short i, j;

  if (argc != 1)
    { if (argc != 2)
        { fprintf(stderr, "Usage: asc2rul3 [<file>]\n");
          return ERR_USAGE; }

      /* argv[1]: source filename */

      if (strlen(argv[1]) > MAXLEN-5)
        { fprintf(stderr, "Error: %s\n", str_too_long);
          return ERR_TOOLNG; }

      s = fname + strlen(strcpy(fname, argv[1]));

      if (!(srce = fopen(fname, "r")))
        { fprintf(stderr, "Cannot open file %s\n", fname);
          return ERR_OPEN; }

      t = s;
      do
        if (s == fname || *s == DS)
          { s = t;
            break; }
      while (*--s != '.');
      strcpy(s, ".rul");

      if (!(dest = fopen(fname, "wb")))
        { fprintf(stderr, "Cannot create file %s\n", fname);
          return ERR_CREATE; }
    }

  for(i=1;i<=NR;i++)
    { strcpy(buffer, residf[i-1]);
      defidf()->value = i; }

  compfile();

  for(i=0;i<14;i++)
    if (!hdrdef[i])
      { fprintf(stderr, "Header not completely defined\n");
        return ERR_HDRDEF; }

  s = buffer;
  p = hdrval + 8;
  *((long *) s)++ = 'RUL3';
  *((long *) s)++ = *p;
  *((long *) s)++ = *--p;
  *((long *) s)++ = *--p;
  *((long *) s)++ = *--p;

  #ifdef SSH2
    #define STO_WORD  *((short *) s)++ = (short) *--p;
    #define STO_WRD(w) *((short *) s)++ = w;
  #else
    #define STO_WORD i = (short) *--p; *s++ = (char) (i >> 8); *s++ = (char) i;
    #define STO_WRD(w) *s++ = (char) ((w) >> 8); *s++ = (char) (w);
  #endif

  STO_WORD

  i = 0;
  while (*s++ = rulname[i++]);
  while (++i <= NAMLEN)  *s++ = '\0';

  STO_WORD
  STO_WORD

  *s++ = (char) npat;
  *s++ = (char) nshp;
  *s++ = (char) nsig;
  *s++ = (char) *--p;

  STO_LONG(*(long *) nsco)

  for(i=0;i<4;i++)
    { if ((unsigned char) s[-4] > hdrval[i+9])
        { fprintf(stderr,
          "The number of the scores is greater than the maximal number (nhex = %d)",
          (int) (i + 6));
          return ERR_SCORES; }
      *s++ = (char) hdrval[i+9]; }

  STO_LONG(*--p)
  STO_LONG(tmp = 110 + 4 * (27 * npat + 2 * nshp + npos))
  STO_LONG(tmp += 8 * nsig)
  tmp += 2 * nnbr + nnrt;

  for(i=0;i<=6;i++)
    { if (paddr[i] < 0)
        { fprintf(stderr, "The part %c is undefined\n", parts[i]);
          return ERR_RULPART; }
      STO_LONG(tmp + paddr[i]) }

  tmp += rullen;
  STO_LONG(textlen ? (textlen += 4, tmp) : 0)
  STO_LONG(tmp + textlen)

  if (!fwrite(buffer, 110, 1, dest))
    { fwrite_error: fprintf(stderr, "Cannot write to the target file\n");
      return ERR_WRITE; }

  if (!npat)
    { fprintf(stderr, "No pattern\n");
      return ERR_NOPATTERN; }

  for(i=0;i<npat;i++)
    if (!fwrite(spat[i] + 1, 108, 1, dest)) goto fwrite_error;

  if (!nshp)
    { fprintf(stderr, "No shape\n");
      return ERR_NOSHAPE; }

  j = 8 * nshp;
  for(i=0;i<nshp;i++)
    { unsigned char *p;
      p = &((Shape *) sshp[i])->nhex;
      s = buffer + 8 * i;
      *s++ = *p++;
      if (*p <= 1)
        { *s++ = *p - 1;
          STO_WRD(0) }
      else
        { short k;
          long *q;
          k = *s++ = *p;
          STO_WRD(j - 8 * i - 2)
          q = &((Shape *) sshp[i])->shape[0];
          do
            { *(long *) (buffer + j) = *q++;
              j += 4; }
          while (--k);
        }
      *(long *) s = ((Shape *) sshp[i])->shape[0];
    }
  if (!fwrite(buffer, j, 1, dest)) goto fwrite_error;

  if (!fwrite(signals, 8 * nsig, 1, dest)) goto fwrite_error;

  s = buffer;
  STO_WRD(nnbr)
  j = 2 * nnbr;

  for(i=0;i<nnbr;i++)
    { STO_WRD(j)
      j += ntlen[i]; }

  if (!fwrite(buffer, 2 + 2 * nnbr, 1, dest)) goto fwrite_error;

  for(i=0;i<nnbr;i++)
    { Textnum *p;
      if (!fwrite(numbers[i].data, 17, 1, dest)) goto fwrite_error;
      p = numbers[i].text;
      do
        { if (!fwrite(&p->len, 2 + p->len, 1, dest)) goto fwrite_error; }
      while (p = p->next);
      if (!fwrite("", 1, 1, dest)) goto fwrite_error;
    }

  while (idl)
    { Idf *p;
      p = idnum;
      while (p)
        { if (!strncmp(idl->idf, p->idf, IDMLEN))
            { tmp = p->value - idl->addr;
              if (idl->size == 1)
                { if (tmp <= 127 && tmp >= -128)
                    { rulcmd[idl->addr] = (char) tmp;
                      goto next_label; }
                }
              else if (tmp <= 32767 && tmp >= -32768)
                {
                  #ifdef SSH2
                    *(short *) (rulcmd + idl->addr) = (short) tmp;
                  #else
                    rulcmd[idl->addr] = (char) (tmp>>8);
                    rulcmd[idl->addr+1] = (char) tmp;
                  #endif
                  goto next_label;
                }
              fprintf(stderr, "Offset related to label %s too high", idl->idf);
              return ERR_OFFSET; }
          p = p->next; }
      fprintf(stderr, "Undefined label %s\n", idl->idf);
      return ERR_UNDEFLBL;
      next_label: idl = idl->next;
    }

  if (!fwrite(rulcmd, rullen, 1, dest)) goto fwrite_error;

  if (hlp)
    { Textblock *t;
      do
        { long ll;
          if (!fwrite(&hlp->len, 4, 1, dest)) goto fwrite_error;
          t = hlp->tb;
          while (t->next)
            { if (!fwrite(&t->text, TBLEN, 1, dest)) goto fwrite_error;
              t = t->next; }
          if ((ll = (hlp->len - 4) & (TBLEN - 1))
           && !fwrite(&t->text, ll, 1, dest))
            goto fwrite_error;
        }
      while (hlp = hlp->next);
      if (!fwrite(&hlp, 4, 1, dest)) goto fwrite_error;
    }

  return 0;
}

void compfile(void)
{
  n = 1;

  /*
   * Compile "srce" file, whose name is "gfn" (gfn == NULL for the
   * main file); "n": current line number of the current file.
   */

  if (!fread(buffer, 9, 1, srce) || strncmp(buffer, "!RUL:ASC1", 9))
    error(ERR_SRCE, "bad source file");

  new_section:
  if (nextc() != '#') error(ERR_SECTION, "new section expected");

  /* New section. The last read character is '#'. */

  for(;;)
    { if (!fread(buffer, 4, 1, srce)) bad_section_name();

      switch (*(long *) buffer)  /* First 4 characters of the section name */
        {

  case 'head': expect("er");

    while (nextc() != '#')
      { short h;
        static char *hdrelem[] = { "rule_name", "stacklen", "nhex", "text_f",
                                   "rule_f", "rule_h", "rule_c", "name_c",
                                   "gamov_c", "pause_c" };
        read_idf2();
        if (strcmp(buffer, "maxnscores"))
          { h = 9;
            do
              if (!strcmp(buffer, hdrelem[h])) goto header_val;
            while (--h >= 0);
            error(ERR_HEADER, "bad header variable name");
          }
        else
          { if (nextc() != '[')
              smns_error:
                error(ERR_SYNTAX, "syntax: maxnscores[<const> | *] = ...");
            if (nextc() == '*')
              h = -1;
            else
              { prevc();
                if ((h = get_num() + 4) < 10 || h > 13)
                  error(ERR_MAXNSC,
                        "\"maxnscores\" index must be between 6 and 9");
              }
            if (nextc() != ']') goto smns_error;
          }
        header_val:
        if (nextc() != '=')
          var_eq_value: error(ERR_SYNTAX, "syntax: <var> = <value>");
        if (h)
          { hdrval[h < 0 ? 9 : h-1] = get_num();
            if (h < 0)
              { hdrval[12] = hdrval[11] = hdrval[10] = hdrval[9];
                hdrdef[13] = hdrdef[12] = hdrdef[11] = hdrdef[10] = -1; }
          }
        else
          get_string(rulname, NAMLEN);
        if (h >= 0) hdrdef[h] = -1; }
    break;

  case 'patt': expect("ern");

    while (nextc() != '#')
      { if (patshp(&pat, spat, sizeof(PatShp)+108, '{', &npat))
          { short i = 288;
            long j = 0;
            char *p;
            p = (char *) (pat + 1);
            do
              { if ((c = nextc() - '0') < 0 || c >= 8)
                  pat_syntax: error(ERR_SYNTAX, "syntax: { [0-7]^288 }");
                j = (j<<3) + c;
                if (!(--i & 7))
                  { *p++ = j>>16;
                    *p++ = j>>8;
                    *p++ = j;
                    j = 0; }
              }
            while (i);
            if (nextc() != '}') goto pat_syntax;
          }
      }
    break;

  case 'shap': expect("e");

    while (nextc() != '#')
      { if (patshp(&shp, sshp, sizeof(Shape), '[', &nshp))
          { short i = 6, j;
            signed char *a, *b;
            static char *turn[7] = { "std", "locked",
                                     "2", "3", "4", "5", "6" };
            nextc();
            a = buffer;
            while (c >= '2' && c <= '6' || c >= 'c' && c <= 't')
              { *a++ = c;
                readc(); }
            *a = '\0';
            prevc();
            do
              if (!strcmp(buffer, turn[i])) goto shape;
            while (--i >= 0);
            turn_error: error(ERR_SYNTAX, "syntax: [ <turn mode> ]");
            shape:
            if (nextc() != ']') goto turn_error;
            b = (char *) shp + sizeof(PatShp);
            if ((j = (b[1] = (char) i) ? i : 1) >= 2)  npos += j;
            a = (char *) ((Shape *) shp)->shape;
            do
              { short k = 18;
                static char hex[19] = { 7, 8, 18, 9, 1, 17, 2, 6, 10, 0,
                                      16, 3, 5, 11, 4, 15, 12, 14, 13 };
                if (nextc() != '{')
                  error(ERR_SYNTAX, "shape position expected");
                a[3] = a[2] = a[1] = a[0] = -1;
                do
                  if (nextc() != '.')
                    { if ((c -= '1') < 0 || c >= 4)
                        error(ERR_SYNTAX, "period or 1, 2, 3 or 4 expected");
                      if (a[c] >= 0)
                        error(ERR_SHP1, "a digit mustn't appear more than once");
                      a[c] = hex[k]; }
                while (--k >= 0);
                nextc();
                err_char('}');
                if (*a++ < 0)
                  shp_err: error(ERR_SHP2, "the shape is not correctly defined");
                if (*a++ < 0)
                  { if (*a++ >= 0 || *a++ >= 0) goto shp_err;
                    *b = 1; }
                else if (*a++ < 0)
                  { if (*a++ >= 0) goto shp_err;
                    *b = 2; }
                else
                  *b = (*a++ < 0) ? 3 : 4;
              }
            while (--j);
          }
      }
    break;

  case 'sign': expect("al");

    while (nextc() != '#')
      { long signame;
        read_idf3();
        if (!(char) ++nsig) error(ERR_TMANYSIG, "too many signals");
        if (!((signame = *(long *) buffer) & 0xFF0000))
          signame &= 0xFFFF;
        else if (!(signame & 0xFF00))
          signame &= 0xFF;
        else if (!(signame & 0xFF) && !buffer[4]) error (ERR_SIGNAME,
                "a signal name mustn't have more than 4 characters");
        signals[nsig-1].name = signame;
        signals[nsig-1].key = get_num() << 16; }
    break;

  case 'numb': expect("er");

    while (nextc() != '#')
      { Textnum *p = NULL;
        static char *nbrelem[] = { "text_pos", "digit_c", "zero_c", "text_c",
                                   "number_h", "text_h", "min", "max" };
        static short offset[] = { 16, 12, 8, 4, 3, 2, 1, 0 };
        static short length[] = { 0, 1, 1, 1, 0, 0, 0, 0 };
        read_idf3();
        if (nextc() != '{')
          error(ERR_SYNTAX, "syntax: <idf>: { ...");
        if (nnbr == 8)
          error(ERR_TMANYNBR, "too many numbers");
        numbers[nnbr].numdef = 0;
        numbers[nnbr].text = NULL;
        ntlen[nnbr] = 18;
        while (nextc() != '}')
          { read_idf2();
            if (strcmp(buffer, "text"))
              { short i = 7;
                long value;
                do
                  if (!strcmp(buffer, nbrelem[i])) goto number_elem;
                while (--i >= 0);
                error(ERR_NUMBER, "bad number variable name");
                number_elem:
                if (nextc() != '=') goto var_eq_value;
                numbers[nnbr].numdef |= 1<<i;
                if (i)
                  value = get_num();
                else
                  { static char *pos[] = { "right", "above", "left" };
                    nextc();
                    read_idf2();
                    value = 2;
                    do
                      if (value < 0) error(ERR_TEXTPOS,
                      "the text position must be \"left\", \"right\" or \"above\"");
                    while (strcmp(buffer, pos[value--]));
                  }
                if (length[i])
                  *(long *) (numbers[nnbr].data + offset[i]) = value;
                else
                  numbers[nnbr].data[offset[i]] = value;
              }
            else
              { if (nextc() != '[')
                  text_error:
                    error(ERR_SYNTAX, "syntax: text[<language>] = ...");
                if (!(p = *(p ? &p->next : &numbers[nnbr].text)
                        = (Textnum *) malloc(sizeof(Textnum))))
                  insuf_mem();
                p->next = NULL;
                p->language = get_language();
                if (nextc() != ']' || nextc() != '=') goto text_error;
                ntlen[nnbr] += (p->len = get_string(p->text, 128)) + 2;
              }
          }
        if (numbers[nnbr].numdef != -1 || !numbers[nnbr].text)
          error(ERR_DEFNBR, "the number is not completely defined");
        nnrt += ntlen[nnbr++]; }
    break;

  case 'rule':

    get_cmd();
    if (c != '#')
      error(ERR_SYNTAX, "command or section header expected");
    break;

  case 'help':

    while (nextc() != '#')
      { char idfhlp[IDMLEN+1];
        if (!(hlplast = *(hlp ? &hlplast->next : &hlp)
                      = (Help *) malloc(sizeof(Help))))
          insuf_mem();
        hlplast->next = hlplast->tb = NULL;
        no_long_idf = -1;
        read_idf3();
        strcpy(idfhlp, buffer);
        no_long_idf = 0;
        thlp = &hlplast->tb;
        helplen = 4;
        add(get_language());
        if (nextc() != ':')
          hlp_error:
            error(ERR_SYNTAX, "syntax: <idf>: <langue> :[\\33-\\255]?\n ...");
        if (readc() <= 32)
          { if (c != '\n') goto hlp_error;
            add(0); }
        else
          { add(c);
            if (readc() != '\n') goto hlp_error;
          }
        n++;
        add(0);
        for(;;)
          { if (readc() == ':')
              { short i = -1, j = 0;
                while (idfhlp[++i])
                  if (readc() != idfhlp[i])
                    { add(':');
                      while (j < i)
                        add(idfhlp[j++]);
                      goto help_cont; }
                break;
              }
            help_cont:
            while (c != '\n')
              { if (c < 32) error(ERR_HELP,
                  "the text mustn't contain control characters except '\\n'");
                add(c);
                readc(); }
            add('\r');
            add('\n');
            n++; }
        add('\0');
        textlen += hlplast->len = helplen; }
    break;

  case 'scor': expect("e");

    while (nextc() != '#')
      { if (!(scolast = *(scores ? &scores->next : &scores)
                      = (Score *) malloc(sizeof(Score))))
          insuf_mem();
        scolast->next = NULL;
        if (scolast->nhex = c != '*' ? (prevc(), get_num()) : 0)
          { if ((scolast->nhex - 6) & ~3) error(ERR_BADNHEX,
                  "the number of hexagones must be between 6 and 9");
            if (!++nsco[scolast->nhex - 6])
              tmany_sco: error(ERR_TMANYSC, "too many scores");
          }
        else
          { short i;
            for(i=0;i<4;i++)
              if (!++nsco[i]) goto tmany_sco;
          }
        if (nextc() != ':')
          error(ERR_SYNTAX, "syntax: (<const> | *) : ...");
        get_string(scolast->name, PLNLEN);
        scolast->scor = get_num();
        scolast->date = get_num(); }
    break;

  case 'defi': expect("ne");

    if (nextc() == EOF)
      error(ERR_SYNTAX, "syntax: #define <idf> ...");
    read_idf2();
    { Idf *p;
      p = defidf();
      p->value = defexpr(1);
      p->var = varflag; }
    goto new_section;

  case 'incl': expect("ude");

    { int oldn;
      short i = 0;
      char fname[MAXLEN], *oldfname;
      FILE *oldsrce;
      if (nextc() == EOF)
        error(ERR_SYNTAX, "syntax: #include <filename>");
      do
        { buffer[i++] = c;
          if (i == MAXLEN)
            error(ERR_TOOLNG, "File name too long");
        }
      while (readc() > 32);
      if (c == EOF)  i = 0;
      oldsrce = srce;
      if (!(srce = fopen(fname, "r")))
        error(ERR_OPEN, "Cannot open the include file");
      oldn = n;
      oldfname = gfn;
      gfn = fname;
      compfile();
      gfn = oldfname;
      n = oldn;
      srce = oldsrce; }
    goto new_section;

  case 'end.': return;

  default: bad_section_name();

        }
    }
}

void get_cmd(void)
{
  for(;;)
    { long r;
      r = get_const(-1, -1);
      if (noconst)
        if (noconst < 0)
          { rulput(RUL_FCT);
            def_idbra(2); }
        else if (varflag)
          { rulput(RUL_VAR);
            if (nextc() != ']' && (r = (prevc(), varexpr(1))) >= 0)
              rulcmd[rullen-1] = (char) r;
          }
        else if (r < 0)
          { if (c != ':') undef_idf(ERR_UNDEFIDF, str_undef);
            nextc();
            defidf()->value = rullen; }
        else if (r)
          if (r >= RETURN)
            { short k;
              rulput(k = idfcod[r-RETURN]);
              if (k >>= 8) rulput(k);
            }
          else if (r >= CHGALL)
            { rulput(RUL_CHG + r - CHGALL);
              nextc();
              err_char('{');
              get_cmd();
              if (c != '}')
                error(ERR_SYNTAX, "command or '}' expected");
              rulput(0xBF);
            }
          else if (r >= TCHG)
            { int orl;
              rulput(RUL_CHG + r - TCHG);
              if (nextc() != '('
               || (rulput(RUL_ADD), orl = rullen, r = infix(1, 0), c != ')'))
                error(ERR_SYNTAX,
                  "syntax: (tchg|tclr|tset) \"(\" <infix> \")\" \"{\" ... \"}\"");
              if (rullen == orl)
                { rullen--;
                  sto_const(r+8); }
              else
                rulput(0xC8);
            }
          else if (r == TIMER_ON)
            { rulput(RUL_TIMER);
              goto var_index; }
          else if (r == GOTO)
            { rulput(RUL_BRA);
              new_idbra(2, 0); }
          else if (r == IFS || r == IFL)
            { rulput(r == IFS ? RUL_IFS : RUL_IFL);
              if (nextc() != '(' || (get_cmd(), c != ')'))
                error(ERR_SYNTAX, "syntax: ( IF | IF.S ) \"(\" ... \")\" <idf>");
              new_idbra(r - (IFS - 1), 0); }
          else
            { long s;
              rulput(r + (RUL_ARRAY16 - ARRW));
              nextc();
              err_char('(');
              get_cmd();
              err_char(',');
              if ((s = get_num()) <= 0 || s >= 256)
                error(ERR_BADSIZE, "the size must be between 1 and 255");
              rulput((char) s);
              nextc();
              err_char(')');
              do
                if (r != CASE)
                  { long cst;
                    cst = get_num();
                    if (r == ARRL)
                      { if (cst < -32768 || cst > 32767)
                          error(ERR_BADCST,
                                "the number must be between -32768 and 32767");
                        rulput((char) (cst >> 24));
                        rulput((char) (cst >> 16)); }
                    rulput((char) (cst >> 8));
                    rulput((char) cst); }
                else
                  new_idbra(2, 2);
              while (--s);
            }
        else if (c == '\\')
          { short i = 2, r = 0;
            do
              { if ((readc() - '0') < 0 ||
                    c > 9 && ((c -= 'A' - '0' - 10) < 10 ||
                              c >= 16 && ((c -= 'a' - 'A') < 10 || c >= 16)))
                  error(ERR_SYNTAX, "syntax: \\<hex><hex>");
                r = (r<<4) + c; }
            while (--i);
            rulput((char) r); }
        else if (c == '=')
          if (readc() == '=')
            rulput(RUL_EQ);
          else
            { prevc();
              rulput(RUL_STO);
              var_index:
              r = get_const(-1, 0);
              if (!varflag)
                var_expected: error(ERR_NOVAR, "variable expected");
              if (!noconst || (r = varexpr(1)) >= 0)
                rulput((char) r + 0xC0);
            }
        else if (c == '+')
          if (readc() == '=')
            { rulput(RUL_STO);
              r = get_const(-1, 0);
              if (!varflag) goto var_expected;
              if (noconst)
                { rulput(RUL_ADD);
                  if ((r = varexpr(1)) >= 0)
                    rulcmd[rullen-1] = (char) r + 0x80;
                  else
                    rulput(0x80);
                }
              else
                rulput((char) r + 0x80);
            }
          else
            { prevc();
              rulput(RUL_ADD); }
        else if (c == '-')
          { if (readc() == '=')
              { rulput(RUL_STO);
                r = get_const(-1, 0);
                if (!varflag) goto var_expected;
                if (noconst)
                  { rulput(RUL_SUB);
                    if ((r = varexpr(1)) >= 0) goto add_sto;
                    rulput(0x80); }
                else
                  { add_sto: rulcmd[rullen-1] = RUL_CST16;
                    rulput(0);
                    rulput((char) r + 0x40); }
              }
            else if (c == '>')
              rulput(RUL_NBOR);
            else
              { prevc();
                if (c == '$' || c >= '0' && c <= '9')
                  sto_const(-get_const(0, 0));
                else
                  rulput(RUL_SUB);
              }
          }
        else if (c == '*')
          rulput(RUL_MUL);
        else if (c == '/')
          rulput(RUL_DIV);
        else if (c == '<')
          if (readc() == '<')
            rulput(RUL_ASL);
          else if (c == '=')
            rulput(RUL_LE);
          else
            { prevc();
              rulput(RUL_LT); }
        else if (c == '>')
          if (readc() == '>')
            rulput(RUL_ASR);
          else if (c == '=')
            rulput(RUL_GE);
          else
            { prevc();
              rulput(RUL_GT); }
        else if (c == '!')
          { if (readc() != '=')
              error(ERR_CMD, "unknown command");
            rulput(RUL_NE); }
        else if (c == '%')
          rulput(RUL_MOD);
        else if (c == '&')
          rulput(RUL_AND);
        else if (c == '|')
          rulput(RUL_OR);
        else if (c == '^')
          rulput(RUL_XOR);
        else if (c == '~')
          rulput(RUL_NOT);
        else if (c == '@')
          { nextc();
            read_idf3();
            if (!*buffer)
              error(ERR_ADDRID, "address identifier expected");
            switch (nextc())
              { case 'B':
                  def_idaddr(1);
                  break;
                case 'W':
                  def_idaddr(2);
                  break;
                case 'L':
                  def_idaddr(4);
                  break;
                case '=':
                  chg_value(0);
                  break;
                case '+':
                  chg_value(1);
                  break;
                case '-':
                  chg_value(-1);
                  break;
                default:
                  error(ERR_SYNTAX, "@ <idf>: B | W | L | (=|+|-) <const>");
              }
          }
        else if (c == '(')
          { int orl;
            orl = rullen;
            r = infix(1, 0);
            if (c != ')') error(ERR_SYNTAX, str_inf);
            if (rullen == orl) sto_const(r);
          }
        else
          break;
      else if (varflag)
        rulput((char) r & 0x3F);
      else
        sto_const(r);
    }
}

void sto_const(long r)
{
  rulput(r >= -64 && r <= 63 ? (char) r + 0xC0 :
    ( rulput(r >= -32768 && r <= 32767 ? RUL_CST16 :
             (rulput(RUL_CST32), rulput((char) (r >> 24)), (char) (r >> 16))),
      rulput((char) (r >> 8)),
      (char) r ));
}

void err_char(char x)
{
  static char s[] = "'\0' expected";

  if (c != x)
    { s[1] = x;
      error(ERR_SYNTAX, s); }
}

void new_idbra(short size, short flag)
{
  nextc();
  read_idf2();
  if (!*buffer) error(ERR_LABEL, "label expected");
  if (flag && buffer[0] == '.' && !buffer[1])
    { rulput(0);
      rulput(0); }
  else
    def_idbra(size);
}

void def_idbra(short size)
{
  Idlabel *p;

  p = idl;
  if (!(idl = (Idlabel *) malloc(sizeof(Idlabel)))) insuf_mem();
  idl->next = p;
  strncpy(idl->idf, buffer, IDMLEN);
  idl->addr = rullen;
  rullen += (idl->size = size);
}

void chg_value(short flag)
{
  Idfaddr *p;

  p = idaddr;
  while (p)
    { if (!strncmp(buffer, p->idf, IDMLEN))
        { long r;
          r = get_num();
          p->value = flag ? p->value + flag * r : r;
          return; }
      p = p->next; }
  undef_idf(ERR_UNDEFADDRID, "undefined address identifier ");
}

void def_idaddr(short size)
{
  Idfaddr *p;

  p = (Idfaddr *) defid0((Idf **) &idaddr, sizeof(Idfaddr));
  p->addr = rullen;
  p->value = 0;
  rullen += (p->size = size);
}

Idf *defid0(Idf **id, int size)
{
  Idf *p;

  if (p = *id)
    { for(;;)
        { if (!strncmp(buffer, p->idf, IDMLEN)) alrdy_def();
          if (!p->next) break;
          p = p->next; }
      id = &(p->next);
    }
  if (!(p = (Idf *) malloc(size))) insuf_mem();
  (*id = p)->next = NULL;
  strncpy(p->idf, buffer, IDMLEN);
  return p;
}

Idf *defidf(void)
{
  Idf *p;

  p = defid0(&idnum, sizeof(Idf));

  if (buffer[0] == '_' && !buffer[2])
    { short k;
      for (k=0;k<7;k++)
        if (buffer[1] == parts[k])
          { paddr[k] = rullen;
            break; }
    }

  p->var = 0;
  return p;
}

void alrdy_def(void)
{
  error(ERR_DEFINED, "Identifier already defined");
}

long defexpr(short flag)
{
  long r, s;

  r = get_const(flag ? -1 : 1, 1);
  if (!noconst)  return r;

  if (varflag)
    { r = nextc() == ']' ? defexpr(0) : (prevc(), varexpr(0));
      varflag = 1;
      return r; }

  if ((short) r)
    { switch ((short) r)
        { case ABS:
            return labs(defexpr(0));
          case SQ:
            r = defexpr(0);
            return r*r;
          case MIN:
            r = defexpr(0);
            s = defexpr(0);
            return r <= s ? r : s;
          case MAX:
            r = defexpr(0);
            s = defexpr(0);
            return r >= s ? r : s; }
      undef_idf(ERR_UNDEFIDF, str_undef);
    }

  if (c == '(')
    { r = infix(0, 1);
      if (c != ')') error(ERR_SYNTAX, str_inf);
      return r; }

  if (c == '+')
    { r = defexpr(0);
      s = defexpr(0);
      return r + s; }

  if (c == '-')
    { short flag;
      flag = nextc() == '$' || c >= '0' && c <= '9';
      prevc();
      r = defexpr(0);
      return flag ? -r : r - defexpr(0); }

  if (c == '*')
    { r = defexpr(0);
      s = defexpr(0);
      return r * s; }

  if (c == '/')
    { r = defexpr(0);
      if (s = defexpr(0))  return r / s;
      error(ERR_DIV0, div0); }

  if (c == '<')
    { if (readc() == '<')
        { r = defexpr(0);
          s = defexpr(0);
          return r << s; }
      if (c == '=')
        { r = defexpr(0);
          s = defexpr(0);
          return r <= s; }
      prevc();
      r = defexpr(0);
      s = defexpr(0);
      return r < s;
    }

  if (c == '>')
    { if (readc() == '>')
        { r = defexpr(0);
          s = defexpr(0);
          return r >> s; }
      if (c == '=')
        { r = defexpr(0);
          s = defexpr(0);
          return r >= s; }
      prevc();
      r = defexpr(0);
      s = defexpr(0);
      return r > s;
    }

  if (c == '%')
    { r = defexpr(0);
      if ((s = defexpr(0)) > 0)  return r >= 0 ? r % s : s - ((-r) % s);
      error(ERR_MOD, mod); }

  if (c == '=')
    { if (readc() == '=')
        { r = defexpr(0);
          s = defexpr(0);
          return r == s; }
    }
  else if (c == '!')
    { if (readc() == '=')
        { r = defexpr(0);
          s = defexpr(0);
          return r != s; }
    }
  else
    { if (c == '&')
        { r = defexpr(0);
          s = defexpr(0);
          return r & s; }
      if (c == '|')
        { r = defexpr(0);
          s = defexpr(0);
          return r | s; }
      if (c == '^')
        { r = defexpr(0);
          s = defexpr(0);
          return r ^ s; }
      if (c == '~')
        { r = defexpr(0);
          return ~r; }
    }

  syntax_error();
}

void syntax_error(void)
{
  error(ERR_SYNTAX, "syntax error");
}

long varexpr(short flag)
{
 /*
  * flag: like infix. If the expression is a constant, the value
  * must be between 0 and 63. If flag = 1 and the expression is
  * not a constant, -1 is returned.
  */

  long r;
  int orl;

  orl = rullen;
  r = infix(flag, 1-flag);
  if (c != ']')
    error(ERR_SYNTAX, "syntax: V[ <expr> ]");
  if (flag && rullen != orl) return -1;
  check_varind(r);
  return r;
}

void check_varind(long r)
{
 if (r < 0 || r >= 64)
   error(ERR_VARIND, "the variable index must be between 0 and 63");
}

long infix(short flag, short d)
{
 /*
  * flag = 0: the expression must be a constant (the value is returned).
  *        1: if the expression is a constant, the value is returned;
  *           otherwise the expression is stored using rulput().
  *       -1: the expression is always stored using rulput().
  */

  long r;
  int orl;
  char op;
  static short t;
  static char bop;

  orl = rullen;

  if (d < 8)
    { long value;
      int orl2;
      short cflag = 0;
      for(;;)
        { orl2 = rullen;
          r = infix(cflag >= 0 ? abs(flag) : -1, d+1);
          if (t != d)
            { if (t < d && orl2 == rullen && (cflag < 0 || flag < 0))
                sto_const(r);
              return r; }
          if (!d)
            { if (!varflag) error(ERR_SYNTAX, var_error);
              if (rullen == orl)
                { check_varind(r);
                  rulput(RUL_STO);
                  sto_const(r + (bop ? (bop == RUL_ADD ? -64 : 64) : 0)); }
              else
                { shift_rul(orl, RUL_STO);
                  if (bop)
                    { shift_rul(orl+1, bop);
                      rulput(0x80); }
                }
              orl = rullen;  /* right assoc. */
              cflag = -1;
            }
          else if (cflag)
            { if (flag)
                { if (orl2 != rullen) goto not_cst1;
                  if (cflag < 0)
                    { sto_const(r);
                      goto not_cst2; }
                  rullen = orl; }
              switch (op)
                { case RUL_ADD:
                    value += r;
                    break;
                  case RUL_SUB:
                    value -= r;
                    break;
                  case RUL_MUL:
                    value *= r;
                    break;
                  case RUL_DIV:
                    if (r)
                      { value /= r;
                        break; }
                    error(ERR_DIV0, div0);
                  case RUL_MOD:
                    if (r > 0)
                      { value = value >= 0 ? value % r : r - ((-value) % r);
                        break; }
                    error(ERR_MOD, mod);
                  case RUL_ASL:
                    value <<= r;
                    break;
                  case RUL_ASR:
                    value >>= r;
                    break;
                  case RUL_EQ:
                    value = value == r;
                    break;
                  case RUL_NE:
                    value = value != r;
                    break;
                  case RUL_GE:
                    value = value >= r;
                    break;
                  case RUL_GT:
                    value = value > r;
                    break;
                  case RUL_LE:
                    value = value <= r;
                    break;
                  case RUL_LT:
                    value = value < r;
                    break;
                  case RUL_OR:
                    value |= r;
                    break;
                  case RUL_XOR:
                    value ^= r;
                    break;
                  case RUL_AND:
                    value &= r;
                    break;
                }
              if (t >= 0) goto cst;
              return value;
            }
          else if (!flag || rullen == orl)
            { cflag = 1;
              value = r;
              cst: op = bop;
              if (flag)
                { rulput(bop);
                  sto_const(value); }
            }
          else
            { not_cst1: cflag = -1;
              not_cst2: shift_rul(orl, bop);
            }
        }
    }

  r = get_const(-1, 1);

  if (noconst)
    if (noconst < 0)
      { set_noconst(flag);
        rulput(RUL_FCT);
        def_idbra(2); }
    else if (varflag)
      { set_noconst(flag);
        r = infix(abs(flag), 0);
        err_char(']');
        varflag = 1; }
    else
      { if (r < 0) undef_idf(ERR_UNDEFIDF, str_undef);
        if (r)
          { if (r == ABS)
              { if (flag) rulput(RUL_ABS);
                r = abs(infix(abs(flag), 8));
                decr_rul: if (rullen - orl == 1) rullen--;
                prevc();
                if (ncop) prevc();
              }
            else if (r == SQ)
              { if (flag) rulput(RUL_SQ);
                r = infix(abs(flag), 8);
                r *= r;
                goto decr_rul; }
            else if (r == MIN)
              { op = RUL_MIN;
                goto min_max; }
            else if (r == MAX)
              { op = RUL_MAX;
                min_max:
                  { short cst = 0;
                    int f = 0, orl3;
                    long value;
                    if (nextc() != '(')
                      { minmax_error:
                        sprintf(buffer,
                                "syntax: m%s ( <expr> , ... , <expr> )",
                                op == RUL_MIN ? "in" : "ax");
                        error(ERR_SYNTAX, buffer); }
                    orl3 = rullen;
                    do
                      { int orl2;
                        orl2 = rullen;
                        r = infix(abs(flag), !flag);
                        if (rullen == orl2)
                          if (cst)
                            if (op == RUL_MIN)
                              { if (r < value)  value = r; }
                            else
                              { if (r > value)  value = r; }
                          else
                            { cst = 1;
                              value = r; }
                        else
                          { if (f) shift_rul(orl3, op);
                            f = 1; }
                      }
                    while (c == ',');
                    if (c != ')') goto minmax_error;
                    if (cst)
                      { r = value;
                        if (f)
                          { shift_rul(orl3, op);
                            sto_const(value); }
                      }
                  }
              }
            else
              { if (!flag) syntax_error();
                switch (r)
                  { case RAND15: case RAND16: case RAND31: case RAND32:
                    case PAUSE: case MSG: case NHEX:
                      rulput(idfcod[r-RETURN]);
                      break;
                    case RADAR:
                      rulput(RUL_NEXT);
                    sto_opp1:
                      rulput(0xBF);
                      break;
                    case CHGALL:
                      rulput(RUL_CHG);
                      goto sto_opp1;
                    case CLRALL:
                      rulput(RUL_CLR);
                      goto sto_opp1;
                    case SETALL:
                      rulput(RUL_SET);
                      goto sto_opp1;
                    case NSHAPE:
                      rulput(RUL_HEX);
                      goto sto_opp1;
                    case HEXRF:
                      rulput(RUL_HEX);
                      rulput(0xC0);
                      break;
                    case COL:
                      op = RUL_COL;
                    insert_op:
                      rulput(op);
                      infix(-1, 8);
                      prevc();
                      if (ncop) prevc();
                      break;
                    case PAT:
                      op = RUL_PAT;
                      goto insert_op;
                    case HEX:
                      op = RUL_HEX;
                      goto insert_op;
                    case NEXT:
                      op = RUL_NEXT;
                      goto insert_op;
                    case X2C:
                      op = RUL_X2C;
                      goto insert_op;
                    case X2H:
                      op = RUL_X2H;
                      goto insert_op;
                    case CH2X:
                      rulput(RUL_CH2X);
                      goto dyad_op;
                    case CHG:
                      rulput(RUL_CHG);
                      goto dyad_op;
                    case CLR:
                      rulput(RUL_CLR);
                      goto dyad_op;
                    case SET:
                      rulput(RUL_SET);
                      goto dyad_op;
                    case TST:
                      rulput(RUL_TST);
                      goto dyad_op;
                    case TCHG:
                      rulput(RUL_CHG);
                    txxx:
                      if (nextc() != '(') goto dyad_error;
                      infix(-1, 0);
                      if (c != ',') goto dyad_error;
                      {
                        int orl;
                        rulput(RUL_ADD);
                        orl = rullen;
                        r = infix(1, 0);
                        if (rullen == orl)
                          { rullen--;
                            sto_const(r+8); }
                        else
                          rulput(0xC8);
                      }
                      goto dyad_end;
                    case TCLR:
                      rulput(RUL_CLR);
                      goto txxx;
                    case TSET:
                      rulput(RUL_SET);
                      goto txxx;
                    default:
                      syntax_error();
                  }
              }
          }
        else if (c == '-')
          if (readc() == '>')
            { rulput(RUL_NBOR);
              dyad_op:
              if (nextc() != '(')
                dyad_error:
                  error(ERR_SYNTAX, "syntax: <dyad2> ( <expr> , <expr> )");
              infix(-1, 0);
              if (c != ',') goto dyad_error;
              infix(-1, 0);
              dyad_end:
              if (c != ')') goto dyad_error;
            }
          else
            { prevc();
              r = - get_num(); }
        else if (c == '~')
          { if (flag) rulput(RUL_NOT);
            r = ~infix(flag, 8);
            goto decr_rul; }
        else
          { if (c != '(') error(ERR_EXPR, "expression expected");
            r = infix(flag, !flag);
            err_char(')'); }
        varflag = 0; }

  ncop = 0;
  switch (nextc())
    { case '=':
        if (readc() == '=')
          { ncop = 1;
            bop = RUL_EQ;
            t = 4;
            break; }
        prevc();
        bop = 0;
        sto_var: t = 0;
        if (varflag) break;
        error(ERR_SYNTAX, var_error);
      case '+':
        bop = RUL_ADD;
        if (readc() == '=')
          { ncop = 1;
            goto sto_var; }
        prevc();
        t = 5;
        break;
      case '-':
        bop = RUL_SUB;
        if (readc() == '=')
          { ncop = 1;
            goto sto_var; }
        prevc();
        t = 5;
        break;
      case '*':
        bop = RUL_MUL;
        t = 6;
        break;
      case '/':
        bop = RUL_DIV;
        t = 6;
        break;
      case '%':
        bop = RUL_MOD;
        t = 6;
        break;
      case '!':
        if (readc() != '=') syntax_error();
        ncop = 1;
        bop = RUL_NE;
        t = 4;
        break;
      case '<':
        if (readc() == '=')
          { ncop = 1;
            bop = RUL_LE;
            t = 4; }
        else if (c == '<')
          { ncop = 1;
            bop = RUL_ASL;
            t = 7; }
        else
          { prevc();
            bop = RUL_LT;
            t = 4; }
        break;
      case '>':
        if (readc() == '=')
          { ncop = 1;
            bop = RUL_GE;
            t = 4; }
        else if (c == '>')
          { ncop = 1;
            bop = RUL_ASR;
            t = 7; }
        else
          { prevc();
            bop = RUL_GT;
            t = 4; }
        break;
      case '&':
        bop = RUL_AND;
        t = 3;
        break;
      case '^':
        bop = RUL_XOR;
        t = 2;
        break;
      case '|':
        bop = RUL_OR;
        t = 1;
        break;
      default:
        if (!varflag && flag < 0 && orl == rullen) sto_const(r);
        t = -1;
    }

  if (varflag && t && rullen == orl)
    { check_varind(r);
      rulput((char) r); }

  return r;
}

void shift_rul(int orl, char x)
{
  int i;
  Idlabel *p;

  rulput(0);
  i = rullen;
  while (--i != orl)
    rulcmd[i] = rulcmd[i-1];
  rulcmd[i] = x;

  p = idl;
  while (p)
    { if (p->addr >= orl) p->addr++;
      p = p->next; }
}

short patshp(PatShp **loc, PatShp **sloc, long size, char next, short *nps)
{
  PatShp *p;
  short selected = 0;
  static char lbuf[10] = { '\0' };

  read_idf3();

  if (nextc() == '<')
    { if (!fread(lbuf, 9, 1, srce) || strncmp(lbuf, "selected>", 9))
        error(ERR_SYNTAX, "syntax: <idf>: <selected> ...");
      selected = 2 * (nextc() == next) - 1; }

  p = *loc;
  while (p)
    { if (!(strncmp(buffer, p->idf, IDMLEN)))
        { if (selected < 0)
            { p->selected = 1;
              addpatshp(nps, p, sloc);
              return 0; }
          selected_error: error(ERR_SELECTED,
              "the conditions about \"<selected>\" are not satisfied");
        }
      p = p->next; }

  if (selected < 0) goto selected_error;
  if (!(p = malloc(size))) insuf_mem();
  p->next = *loc;
  if ((*loc = p)->selected = selected)  addpatshp(nps, p, sloc);
  strncpy(p->idf, buffer, IDMLEN);
  return 1;
}

void addpatshp(short *nps, PatShp *p, PatShp **sloc)
{
  sloc[*nps] = p;
  defidf()->value = ++*nps;
  if (!(char) *nps)
    error(ERR_TMANYPS, "Too many patterns or shapes");
}

short get_string(char *s, short m)
{
  short h = 0;

  if (nextc() != '\"')
    string_error: error(ERR_STRING, "bad string syntax");

  while (readc() != '\"')
    { if (c == '\\') readc();
      if (c < 32) goto string_error;
      if (++h == m) error(ERR_STRLONG, str_too_long);
      *s++ = c; }

  *s = '\0';
  return h;
}

long get_const(short b, short flag)
{
 /*
  *  Get a constant number.
  *
  *  b = -1: A variable (identifier or V[...]) is accepted. 6 cases:
  *          _ Constant number (cf b = 1): the number is returned,
  *            noconst = 0, varflag = 0.
  *          _ Reserved identifier: the rank (between 1 and NR included)
  *            of the identifier is returned, noconst = 1, varflag = 0.
  *          _ Number identifier: the number is returned,
  *            noconst = 0, varflag = 0.
  *          _ Variable identifier: the index of the variable is returned,
  *            noconst = 0, varflag = 1.
  *          _ V[...]: 0 is returned (the index is not evaluated by
  *            this function), noconst = 1, varflag = 1.
  *          _ Undefined identifier: -1 is returned,
  *            noconst = 1, varflag = 0.
  *          _ Function call (address identifier followed by "()"), only
  *            if flag != 0: 0 is returned, noconst = -1, varflag = 0.
  *          _ Nothing: 0 is returned,
  *            noconst = 1, varflag = 0.
  *       0: A language identifier (2 or 3 letters) is accepted. 2 cases:
  *          _ Case like b = 1: the number is returned.
  *          _ Language identifier: the corresponding number is returned.
  *       1: Variables are not accepted (--> varflag = 0).
  *
  *  flag = -1, 0 or 1.
  *  _ If flag = 0, noconst = 1 and varflag = 0, then an error is returned.
  *  _ If flag = -1 and if it is a defined (or reserved) identifier
  *    followed by ':', then an error is returned.
  */

  long r = 0;

  varflag = noconst = 0;
  if (nextc() == '$')
    { short d;
      while ( (d = readc() - '0') >= 0 && d <= 9
           || (d += '0' - 'A' + 10) >= 10 && d < 16 )
        r = (r<<4) + d;
      prevc(); }
  else if (c >= '0' && c <= '9')
    { do
        r = 10 * r + (c - '0');
      while (readc() >= '0' && c <= '9');
      prevc(); }
  else if (!flag && c == '-')
    { r = get_const(b, 0);
      if (varflag)
        error(ERR_SYNTAX, "syntax: - <number>");
    }
  else if (!flag && c == '(')
    { r = infix(0, 1);
      if (c != ')') error(ERR_SYNTAX, str_inf);
    }
  else
    { read_idf1(0);
      if (*buffer)
        { Idf *p;
          short i = 0;
          if (flag && c == '(')
            { if (readc() == ')')
                { noconst = -1;
                  return 0; }
              prevc(); }
          prevc();
          if (!b)
            { static char *languages[17] = { "us", "de", "fr", "uk", "es",
                                         "it", "se", "chf", "chd", "tu", "fi",
                                         "no", "dk", "sa", "nl", "cz", "hu" };
              do
                if (!strcmp(buffer, languages[i]))  return i;
              while (++i != 17);
              i = 0; }
          p = idnum;
          while (p)
            { if (!strncmp(buffer, p->idf, IDMLEN))
                { if (flag < 0 && c == ':')
                    alrdy_def();
                  if (!(varflag = p->var) || b < 0)
                    { if (i < NR) set_noconst(flag);
                      return p->value; }
                  error(ERR_VAR, "variable identifiers are not allowed");
                }
              i++;
              p = p->next; }
          noconst = 1;
          if (b < 0 && buffer[0] == 'V' && !buffer[1] && c == '[')
            { readc();
              varflag = 1;
              return 0; }
          if (flag)
            return -1;
          undef_idf(ERR_UNDEFIDF, str_undef); }
      set_noconst(flag); }

  return r;
}

void set_noconst(short flag)
{
  if (!flag)
    error(ERR_NOCONST, "constant number expected");
  noconst = 1;
}

void undef_idf(int err, char *s)
{
  char t[30+IDMLEN];

  error(err, strncat(strcpy(t, s), buffer, IDMLEN));
}

void read_idf1(short flag)
{
  short i = IDMLEN;
  char *s;

  prevc();
  s = buffer;
  if (c < '0' || c > '9')
    while ( readc() == '_'  ||  c == '.'  ||  c == '?'
        ||  c >= '0' && c <= '9'  ||  (signed char) c < 0
        ||  c >= 'A' && c <= 'Z'  ||  c >= 'a' && c <= 'z' )
      if (i)
        { --i;  *s++ = c; }
      else if (no_long_idf)
        error(ERR_LONGIDF, "identifier or number too long");
  if (flag && s == buffer)
    error(ERR_NOIDF, "Identifier expected");
  *s = '\0';
}

void read_idf2(void)
{
  read_idf1(1);
  prevc();
}

void read_idf3(void)
{
  read_idf1(1);
  if (c != ':')
    error(ERR_DEFIDF, "the identifier must be followed by ':'");
}

void error(int err, char *s)
{
  fprintf(stderr, gfn ? "File %s" : "Main file", gfn);
  fprintf(stderr, ", line %d: %s\n", n, s);
  exit(err);
}

void insuf_mem(void)
{
  error(ERR_MEM, "insufficient memory");
}

void bad_section_name(void)
{
  error(ERR_SECTNAME, "bad section name");
}

void expect(char *s)
{
  int r;

  if (!fread(buffer, r = strlen(s), 1, srce) || strncmp(buffer, s, r))
    bad_section_name();
}

int readc(void)
{
  if (pflag && --pflag)
    c = c2;
  else
    { c2 = c;
      while ((c = fgetc(srce)) == '\r') ;
      if (c == EOF) error(ERR_EOF, "EOF");
    }
  if (c == '\n') n++;
  return c;
}

int nextc(void)
{
  do
    if (readc() == ';') while (readc() != '\n');
  while ((unsigned int) c <= 32);
  return c;
}

void prevc(void)
{
  if (++pflag == 2)
    c = c2;
  else if (ungetc(c, srce) == EOF)
    error(ERR_UNGETC, "\"ungetc\" error");
  if (c == '\n') n--;
}

void add(char x)
{
  if (!*thlp)
    { if (!(*thlp = (Textblock *) malloc(sizeof(Textblock)))) insuf_mem();
      (*thlp)->next = NULL;
      nhlp = 0; }

  (*thlp)->text[nhlp] = x;
  helplen++;

  if (++nhlp == TBLEN)
    thlp = &(*thlp)->next;
}

void rulput(char x)
{
  if (rullen == rulsize && (!(rulcmd = realloc(rulcmd, rulsize += RCMD_ADD))))
    insuf_mem();
  rulcmd[rullen++] = x;
}
